/****************************************************************************
*           K E R N E L   M O D E   J O Y S T I C K   D R I V E R           *  
*---------------------------------------------------------------------------*
* Written by:   Gary T. Desrosiers and Kenneth W. Clapp                     *
* Date:         May 12th, 1996                                              *
* Copyright (c) Gary T. Desrosiers, 1996. All rights reserved               *
* Copyright (c) Kenneth W. Clapp, 1996. All rights reserved                 *
*  																			*
* Description:  Implements the read of the joystick resistive loads and     *
*               button states. This is implemented in a ReadFile function   *
*               where the buffer is actually a pointer to a JOYINFO struct  *
*               and the read length is actually the joystick number to read *
*               + 1. JOYSTICKID1+1 reads the first joystick and             *
*               JOYSTICKID2+1 reads the second.                             *
****************************************************************************/

#include "ntddk.h"
#include "devioctl.h"
#include "joyio.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT,DriverEntry)
#pragma alloc_text(INIT,CreateDevices)
#endif


NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,IN PUNICODE_STRING RegistryPath) {
NTSTATUS status;

	status = CreateDevices(DriverObject,RegistryPath);
    DriverObject->MajorFunction[IRP_MJ_CREATE] = joyioOpen;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = joyioClose;
    DriverObject->MajorFunction[IRP_MJ_READ] = joyioRead;
    status = STATUS_SUCCESS;
    return(status);
}


NTSTATUS joyioOpen(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp) {  
NTSTATUS status;
PDEVICE_EXTENSION deviceExtension;

	deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
	status = STATUS_SUCCESS;
    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return(status);
}


NTSTATUS joyioClose(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp) {
NTSTATUS status;
PDEVICE_EXTENSION deviceExtension;

	deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
    status = STATUS_SUCCESS;
    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	 return(status);
}


NTSTATUS joyioRead(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp) {  
NTSTATUS status;
PJOYINFO pji;
int a_x, a_y, a_b;
ULONG joyID;
PDEVICE_EXTENSION deviceExtension;

   	deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
	pji = (PJOYINFO)Irp->AssociatedIrp.SystemBuffer;
    joyID =  (Irp->Tail.Overlay.CurrentStackLocation->Parameters.Read.Length)-1;

_asm {
	push    eax                ; save registers
	push    ebx
	push    ecx
	push    edx
	mov     a_x,0              ; set the 
	mov     a_y,0              ; starting values  
	mov     ecx,5000           ; runaway counter          
	xor     al,al              ; trigger value            
	mov     dx,0201h           ; game port address 
    cli          		       ; needed to ensure consistency 
	out     dx,al              ; trigger game port event  
    in      al,dx              ; read game port bits      
    push    ax                 ; save for later           
rxyloop:
	mov     bx,cx              ; save the counter
	mov     cl,byte ptr[joyID] ; joystick number to read
	shl     cl,1               ; compute shift right value
	shr     al,cl              ; position values based on JOYSTICKID
	mov     cx,bx              ; restore counter
    and     al,03h             ; mask off JOYSTICKID specified
	test    al,00fh            ; anything left to do ?                     
	jz      rxydone            ; no... return values     
	shr     al,1               ; get ax bit in carry flag 
	adc     a_x,0
	shr     al,1               ; get ay bit in carry flag   
	adc     a_y,0
    in      al,dx              ; read game port bits     
	loopnz  rxyloop
rxydone:
	sti                        ; match the previous cli 
    pop    ax                  ; get the saved port bits
	mov    cl,byte ptr[joyID]  ; compute the
	shl    cl,1                ; shift count again
	shr    al,4                ; isolate bits
	shr    al,cl               ; joystick 1 or 2 buttons
	xor    al,0fh              ; It uses inverse logic
	mov    a_b,0               ; clear first
	mov    byte ptr a_b,al     ; save button bits
	pop    edx                 ; restore registers
	pop    ecx
	pop    ebx
	pop    eax
}
 	pji->wXpos = a_x;
	pji->wYpos = a_y;
	pji->wZpos = 0;
    pji->wButtons = a_b;	
	status = STATUS_SUCCESS;
	Irp->IoStatus.Status =  status;
	Irp->IoStatus.Information = sizeof(JOYINFO);
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return(status);
}


NTSTATUS CreateDevices(IN PDRIVER_OBJECT DriverObject,PUNICODE_STRING RegistryPath) {
NTSTATUS status;
UNICODE_STRING usName;   
UNICODE_STRING usLinkName;
PDEVICE_OBJECT deviceObject[1];

	RtlInitUnicodeString(&usName,L"\\Device\\joyio");
	status = IoCreateDevice(DriverObject,
							sizeof(DEVICE_EXTENSION),
							&usName,
							FILE_DEVICE_UNKNOWN,
							0,      
				            FALSE,
				            &deviceObject[0]);
	if(!NT_SUCCESS(status));         
        deviceObject[0]->Flags |= DO_BUFFERED_IO;
	RtlInitUnicodeString (&usLinkName,L"\\DosDevices\\joyio");
	status = IoCreateSymbolicLink (&usLinkName,&usName);
	return (status);
}


